SOAP.logical <-
function(x)
{
  val <- as.logical(x)

  if(is.na(val)) {
   val <- as.logical(as.integer(x))
  }

  val
}  

SOAPPrimitiveConverters <-
                list("xsd:timeInstant" = as.POSIXct,
                     "xsd:float" = as.numeric,
                     "xsd:int" = as.integer,
                     "xsd:float" = as.numeric,
                     "xsd:string" = as.character,
                     "xsd:boolean" = SOAP.logical,
# Extensions from Datatypes schema                     
                     "xsd:decimal" = as.numeric)


getNodeById <-
  #
  # Find the top-level node identified by id=value of id
  # Strip the `#' prefix off `id' first.
function(id, root)
{
  id <- gsub("^#", "", id)

  for(i in xmlChildren(root)) {
    k <- xmlAttrs(i)
    if(!is.null(k) && !is.na(match("id", names(k))))
      if(id == k[["id"]])
        return(i)
  }

  return(NULL)
}


# Also need general converters that work on nodes
setGeneric("fromSOAP", function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE) standardGeneric("fromSOAP"))

setMethod("fromSOAP", c("character", type="ArrayType"),
 function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE)          
 {
    root <- parseSOAP(node)
    if(xmlName(root) == "Body")
        node <- root[[1]]
    else
        node <- root

    fromSOAPArray(node, type, root, converters)   
 })

setMethod("fromSOAP", c("character", "PrimitiveSOAPType"),
          function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE) {
             node <- parseSOAP(node, asText = TRUE)
             fromSOAP(node[[1]])
          })

setMethod("fromSOAP", "character",
          function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE) 
          {
               root <- parseSOAP(node)
               if(xmlName(root) == "Body")
                 node <- root[[1]]
               else
                 node <- root

               fromSOAP(node, root, converters, append, type)
          })
  # Top-level entry point to convert a top-level SOAP XML node to an R object.
  #
fromSOAP.default = 
function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE)
{
  if(!missing(converters) && append) {
    SOAPPrimitiveConverters[names(converters)] <- converters
    converters <- SOAPPrimitiveConverters
  }

  if(findReturn) {
k = node    
   while(!is.null(node) && xmlName(node) != "return") {
    node = node[[1]]
   }

   if(is.null(node)) {
    node = k
#    stop("Can't find return node")
   }
  }
  

  a <- xmlAttrs(node)
  
  if(!is.null(a)) {
    if(!is.na(match("null", names(a))))
      return(NULL)

    if(is.null(type) && !is.na(match("type", names(a))))
      type <- a[["type"]]

    if(!is.na(match("href", names(a)))) {
      n <- getNodeById(a[["href"]], root)
      if(is.null(n))
        stop("Can't find element ", a[["href"]])
      return(fromSOAP(n, root = root, converters = converters))
    }
  }

  #XX browser()
  if(xmlName(node) == "Array" || (!is.null(type) && is.character(type) && type == "SOAP-ENC:Array")) {
    return(fromSOAPArray(node, root = root, converters = converters, type = NULL)) # type = type
  } else if(xmlSize(node) > 1) {
    return(fromSOAPStruct(node, root = root, converters = converters, type = type))
  } 

  if(is.null(type)) {
     # Want to check namespace is SOAP-ENC
     # and ideally don't want to prefix it with xsd:
     # but probably the most usual case is to get the typ
     # as xsd:type from the attribute and so that is the
     # default.
    type <- paste("xsd", xmlName(node), sep=":")
  }
  
  which <- match(type, names(converters))
  if(!is.na(which))
    val <- converters[[which]](xmlValue(node))
  else {
    val <- xmlValue(node)
    warning("Don't understand the SOAP type `", type, "' yet")
  }
#XX browser()
  val
}
setMethod("fromSOAP", c(type="missing"), fromSOAP.default)
setMethod("fromSOAP", c(type="NULL"), fromSOAP.default)
setMethod("fromSOAP", c(type="character"), fromSOAP.default)

setMethod("fromSOAP", c(type="ANY", "ANY", "ANY"), fromSOAP.default)

setMethod("fromSOAP", c(type="ArrayType"),
function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE)
{
  fromSOAPArray(node, type, root, converters)
})

setMethod("fromSOAP", c(type="ClassDefinition"),
function(node, root = NULL, converters = SOAPPrimitiveConverters, append = TRUE, type = NULL, findReturn = TRUE)
{
  # Need to handle name here to get into S style?
 newSOAPClass(node, type@name, converters, type = type)
})

fromSOAPArray <-
  # Need to handle the partial arrays
  # where individual elements are  specified.
  #
function(node, type = NULL, root = NULL, converters = SOAPPrimitiveConverters)
{
   # Get to the return value. 
  if(xmlName(node[[1]]) == "return")
    node = node[[1]]
  
  len <- xmlSize(node)
  a <- xmlAttrs(node)  
  
  if(is.null(type)) {
    type <- a[["arrayType"]]
    if(!is.null(type)) {
      els <- strsplit(type, ":")[[1]]
      type <- gsub("\(.*\)\\\[.*", "\\1", els[2])
      len <- as.integer(gsub(".*\\\[\([0-9,]+\)\\\]", "\\1", els[2]))
      cat("Type:", type, " & length = ", len, "\n")
    }
  }

  if(is(type, "ArrayType")) {
    type = type@elType
  }
  

   # This is the general mechanism for dealing with offsets.
   # It doesn't work in S-Plus since it uses lexical scoping.
   # Need an OOP object for that.
  offset <- 1 
  if(!is.na(match("offset", names(a)))) {
    tmp <- gsub("\\\[\([0-9]+\)\\\]", "\\1", a[["offset"]])
    offset <- as.integer(tmp)
  }

  ans <- vector("list", len)
  val <- xmlApply(node, function(x, type = NULL, root = NULL, converters = SOAPPrimitiveConverters) {
                           z <- fromSOAP(x, root = root, converters = converters, findReturn = FALSE, type = type)

                           a <- xmlAttrs(x)
                           if(!is.null(a) && !is.na(match("position", names(a)))) {
                             offset <- as.integer(a[["position"]]) + 1
                           }                           

                           ans[[offset]] <<- z
                            # Do we need to differentiate between this coming from
                            # the position or the global offset.
                            # In the case of position, each item will provide 
                            # its own position value.
                           offset <<- offset + 1
                           z
                        }, type=type, converters = converters, root = root)

  return(ans)
}

fromSOAPStruct <-
function(node, root = root, converters = SOAPPrimitiveConverters, type = NULL)
{
   # See if there is a type="value" attribute which we will use
   # for the class.
  a <- xmlAttrs(node)  
  if(!is.null(a) && !is.na(match("type", names(a)))) {
    typeName <- a[["type"]]
  } else
    typeName <- xmlName(node)

  # Now lookup the converters to see if there is an appropriate
  # handler for this type.
#XXX

  # Otherwise, just use the default mechanism.
  val <- xmlApply(node, fromSOAP, root = root, converters = converters, findReturn = FALSE)

  
  class(val) <- gsub("^[a-zA-Z]+:", "", typeName)
  
  val
}  
